#!/usr/bin/python3

"""This is a script for converting the output of a Digital Eksamen
questionnaire to grades on the seven step scale. Specifically, the
questionnaire is assumed to be of the type "Automatic" and to be
composed of Basic Questions and (two-column) Matrix Questions.

NOTE: NO guarantee of correctness is implied. Do perform sanity checks
of the output (e.g check the grades calculated against manually
calculated valued for some randomly selected students).

The point calculation method used is as described now [many others can
be envisioned - the relevant functions to adapt in the program are
findmatrixQpercent(r,i,j) and def findsingleQpercent(r,i).]:

Let p be the maximum points designated for a given question. For a Basic
Question, a correct answer gives p points, a wrong answer or no
answer gives zero points. For a Matrix Question, let k be the number of
rows. For a row where the correct column is chosen, k/p points counted
For a row where the wrong column is chosen, -k/p points are
counted. For a row where no column is chosen, 0 points are
counted. The total points for the question are the sum of these row
counts, *except* that a negative sum is converted to zero points.

How to use the script:

In Digital Eksamen find the exam in question. In the menu Go to MCQ
overview, click on See hand-in in MCQ module. On the resulting page,
click the button marked Export at the bottom. This gives an .xlsx file
with all answers. Open in e.g. Libreoffice or Excel and save the data
as csv [specifics in Libreoffice: use menu point File->Save as->csv,
ignore warning/confirm csv, accept suggested settings of utf-8 and ;
as separator].

Then adapt the global constants in section "Input Specification" below
to the specifics of the data sheet and the exam. Most easily done with
the .xlsx-file open. Also consider doing a few sanity checks via some
of the commented-out print() statements scattered across the program.

Then run the script on the csv-file

  getGrades.py answerFile.csv

The output is the resulting grades shown in several ways. The grading
process consists of adapting the grading scale (the global constant
"gradeStarts") to the viewpoints of the examiner and the external
examiner (e.g. based on their impressions of how the exam questions
worked out in practice).

Save the output (redirection in the shell) in a txt-file.

Note: The current file is a sanitized version of that used in the
Spring of 2023 in the course DM507/DM578/DS814. The data (anonymized)
from that exam and the exam set itself (pdf of the Latex source) are
included with this script. If wanted, these can be used to check out
the process.

"""

import sys
import bisect
from collections import Counter

############################ Grading Scale ##########################

grades = ["-3","00","02","4","7","10","12"]
# gradeStarts states first percentage value giving each grade (except
# lowest grade, of course)
#
gradeStarts = [15,50,56,66,81,91]
#
# Above scale is what we (examiner and external examiner) have used as
# starting point the last years in DM507. It is very close to the
# rating scale suggested by the math censor chairman when the seven
# step scale was introduced in 2006. This scale
# http://www.imada.sdu.dk/~hjm/knuthskala.efterfoelger.html is
# gradeStarts = [15,50,55,66,80,91]

# Final scale that we ended up using F20:
# gradeStarts = [15,47,56,66,81,91]

# Final scale that we ended up using F22:
# gradeStarts = [25,47,54,66,82,92]

# Final scale that we ended up using F23:
gradeStarts = [25,46,54,66,81,92]

# Added to percent before converting to grade, to ensure that any
# deviation between float value and mathematically true value does not
# give a grade one step too low:
epsilon = 10**(-6)

######################### Input Specification ##########################

# For ease, row and column IDs are expressed in spreadsheet numbering
# (1,2,... for rows and A, B,...,AA,.. for columns) and then converted
# to Python 0-indexed indices.
def convertrow(i):
    return i-1
def convertcol(j):
    sum = 0
    for l in range(len(j)):
        sum = 26 * sum + (ord(j[l]) - ord("A") + 1)
    return sum-1

# Global constants, change for each spreadsheet.
# These are for DM507/DM578/DS814 F23:
separator = ";"
studentIDcol = convertcol("A")
questionrow = convertrow(2)
correctanswerrow = convertrow(6)
questionfirstcol = convertcol("F")
questionlastcol = convertcol("BV")
firststudentrow = convertrow(9)
laststudentrow = convertrow(129)
pointsForQs = [2,2,2,2,5,5,3,3,2,3,6,4,3,6,5,2,4,4,4,6,3,4,2,2,2,2,2,2,2,2,2,2]

#Sanity check:
#print(len(pointsForQs))
#print(sum(pointsForQs))

############################ actual program ##########################

def getGrade(percent):
    return grades[bisect.bisect(gradeStarts,percent+epsilon)]

inputcsvfilename = sys.argv[1]
with open(inputcsvfilename, 'r') as f:
    rows = [l.strip().split(separator) for l in f.readlines()]

def findQstarts(QIDs):
    Qstarts = []
    previousQnumber = None
    for i in range(len(QIDs)):
        Qnumber =  QIDs[i].split(".")[0] # leading part of ID (or ID, if no ".")
        if Qnumber != previousQnumber:
# Next line can be temporarily uncommented to find the indices of
# questions, for possible special handling in findXXXQpercent(). For
# Spring 23, this was necessary for question 8, due to an error during
# manual upload (wrong question type chosen). See *) below. We keep
# the code here to save time in case similar situations appear.
#            print(f"Spørgsmål {Qnumber} har i = {i}")
            Qstarts.append(i)
            previousQnumber = Qnumber
    return Qstarts

QIDs = rows[questionrow][questionfirstcol:questionlastcol+1]
correctanswers = rows[correctanswerrow][questionfirstcol:questionlastcol+1]
#Sanity check:
#print(QIDs)
#print(correctanswers)

Qstarts = findQstarts(QIDs)
QstartsExtended = Qstarts + [len(QIDs)]
#Sanity check:
#print(Qstarts)
#print(QstartsExtended)

def findsingleQpercent(r,i):
    qi = questionfirstcol+i
    correct = rows[correctanswerrow][qi]
    student = rows[r][qi]
    return 1.0 if student == correct else 0.0
# *) Next lines substituted the (single) line above in Spring 23:
#    if i != 19:
#        return 1.0 if student == correct else 0.0
#    else: # special handling of question 8 in F23 (Q8 has index i = 19).
#        correct = ["3","4","5"]
#        return 1.0 if student in correct else 0.0
    
def findmatrixQpercent(r,i,j):
    percent = 0.0
    percentstep = 1.0/float(j-i)
    for k in range(i,j):
        qk = questionfirstcol+k
        correct = rows[correctanswerrow][qk]
        student = rows[r][qk]
        percent += (percentstep if student == correct else 0.0 if student == "" else -percentstep)
    return max(0.0,percent)

def findQpercentsforstudent(rownumber):
    percents = []
    studentanswers = rows[rownumber][questionfirstcol:questionlastcol+1]
    for i in range(len(Qstarts)):
        Qstart = QstartsExtended[i]
        nextQstart = QstartsExtended[i+1]
        Qsize = nextQstart - Qstart
        if Qsize == 1:
            percents.append(findsingleQpercent(rownumber,Qstart))
        elif Qsize > 1:
            percents.append(findmatrixQpercent(rownumber,Qstart,nextQstart))
        else:
            print("Error in findPointsForStudent, Qsize < 1!")
    return percents

#Sanity check:
#To see details of a specific student, to do manual checks of the
#program output:
#print(findQpercentsforstudent(78))
#sys.exit()

stat = Counter()
resultList = []                        

print("\nSorted on original row order (points normalized to 100.0 as max):")
for row in range(firststudentrow,laststudentrow+1):
    studentID = rows[row][studentIDcol]
    percents = findQpercentsforstudent(row)
    totalpoints = sum([i*j for (i,j) in zip(pointsForQs,percents)])
    pts = 100.0*totalpoints/sum(pointsForQs)
    grade = getGrade(pts)
    stat[grade] += 1
    print(f"{studentID:<9}", f"{grade:>2}", f"{pts:8.4f}")
    resultList.append((pts,grade,studentID,percents))
    
print("\nSorted on points (normalized to 100.0 as max):")
for (pts,grade,studentID,percents) in sorted(resultList):
    print(f"{studentID:<9}", f"{grade:>2}", f"{pts:8.4f}")

print("\nSorted on points (normalized to 100.0 as max), now with details of each question (correctness in percents):")
for (pts,grade,studentID,percents) in sorted(resultList):
    print(f"{studentID:<9}", f"{grade:>2}", f"{pts:8.4f}", *percents)

print("\nStatistics:")
total = stat.total()
for g in grades:
    print(f"{g:>2}: {stat[g]:>2} {100*stat[g]/total:5.2f}%")
